from gunicorn.app.base import BaseApplication


class _GunicornApplication(BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        for key, value in self.options.items():
            self.cfg.set(key, value)

    def load(self):
        return self.application


def run_wsgi_app(app, address, key_path=None, cert_path=None, workers=4):
    options = {"bind": address, "workers": workers, "worker_class": "sync" if workers == 1 else "gthread"}
    if key_path and cert_path:
        options["certfile"] = str(cert_path)
        options["keyfile"] = str(key_path)
    _GunicornApplication(app, options).run()



import json
from pathlib import Path
from tqdm import tqdm
from contextlib import contextmanager
import threading
import concurrent.futures


def jsonl_reader(path: Path):
    with path.open(encoding="utf-8") as fi:
        for line in tqdm(fi):
            yield json.loads(line)


def dump_line(data):
    return json.dumps(data, ensure_ascii=False)


@contextmanager
def jsonl_writer(path: Path):
    write_lock = threading.Lock()
    with path.open("a", encoding="utf-8") as fo:

        def _writer(data):
            with write_lock:
                fo.write(dump_line(data) + "\n")

        yield _writer


def jsonl_write_list(path: Path, lst):
    with path.open("a", encoding="utf-8") as fo:
        for obj in lst:
            fo.write(dump_line(obj) + "\n")


def jsonl_map(input_path: Path, output_path: Path, func, num_threads=32):
    futures = []

    # read through output first to find missing jobs
    if not output_path.is_file():
        output_path.touch()

    completed_obj_ids = set()
    for data in jsonl_reader(output_path):
        if "id" in data and "output" in data and data["output"]:
            completed_obj_ids |= {data["id"]}

    def process_obj(obj):
        obj["output"] = func(obj)
        return obj

    with jsonl_writer(output_path) as write:
        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
            for obj in jsonl_reader(input_path):
                if "id" not in obj or obj["id"] not in completed_obj_ids:
                    futures.append(executor.submit(process_obj, obj))
            for future in tqdm(concurrent.futures.as_completed(futures), total=len(futures), desc="Mapping over jsonl"):
                result = future.result()
                if result is not None and not isinstance(result, Exception):
                    write(result)



import http.cookies
from datetime import datetime, timezone
from email.utils import formatdate


def now():
    return datetime.now(timezone.utc)


def http_date(dt=None):
    """
    Format the datetime for HTTP cookies
    """
    if not dt:
        dt = now()
    return formatdate(dt.timestamp(), usegmt=True)


def create_cookie(cookie_name, cookie_domain, token, expiry):
    cookie = http.cookies.Morsel()
    cookie.set(cookie_name, token, token)
    # tell the browser when to stop sending the cookie
    cookie["expires"] = http_date(expiry)
    # restrict to our domain, note if there's no domain, it won't include subdomains
    cookie["domain"] = cookie_domain
    # path so that it's accessible for all API requests, otherwise defaults to not the whole site
    cookie["path"] = "/"
    if cookie_domain == "localhost":
        # send only on requests from first-party domains
        cookie["samesite"] = "Strict"
    else:
        # send on all requests, requires Secure
        cookie["samesite"] = "None"
        # only set cookie on HTTPS sites in production
        cookie["secure"] = True
    # not accessible from javascript
    cookie["httponly"] = True

    return cookie.OutputString()


def parse_cookie(cookie_name, http_cookie_str):
    cookie = http.cookies.SimpleCookie(http_cookie_str).get(cookie_name)

    if not cookie:
        return None

    return cookie.value



# LLM utils

import json

from openai import OpenAI
from os import environ


def tool(name, description, parameters):
    def decorator(fn):
        fn.tool_name = name
        fn.tool_definition = {
            "type": "function",
            "name": name,
            "description": description,
            "parameters": parameters,
        }
        return fn
    return decorator


class LLM:
    def __init__(self, api_key=None, base_url=None, model=None, client=None, **kwargs):
        self.api_key = api_key or environ.get("LLM_API_KEY")
        if not self.api_key:
            raise ValueError("No LLM API key found")
        self.base_url = base_url or environ.get("LLM_BASE_URL") or "https://px.axess-ai.com/v1"
        self.model = model
        self.client = client or OpenAI(base_url=self.base_url, api_key=self.api_key, **kwargs)

    def with_model(self, model):
        return LLM(api_key=self.api_key, base_url=self.base_url, model=model, client=self.client)

    def complete(self, user_prompt, system_prompt=None, model=None, **kwargs):
        use_model = model or self.model
        if not use_model:
            raise ValueError("Model not provided in constructor nor at completion")
        response = self.client.responses.create(
            model=use_model,
            input=user_prompt,
            **({"instructions": system_prompt} if system_prompt else {}),
            **kwargs,
        )
        return response.output_text

    def json_complete(self, user_prompt, response_format, system_prompt=None, model=None, **kwargs):
        use_model = model or self.model
        if not use_model:
            raise ValueError("Model not provided in constructor nor at completion")
        response = self.client.responses.parse(
            model=use_model,
            input=user_prompt,
            text_format=response_format,
            **({"instructions": system_prompt} if system_prompt else {}),
            **kwargs,
        )
        return response.output_parsed.model_dump()

    def tool_complete(self, user_prompt, tools, system_prompt=None, model=None, max_rounds=10, **kwargs):
        use_model = model or self.model
        if not use_model:
            raise ValueError("Model not provided in constructor nor at completion")
        tool_map = {t.tool_name: t for t in tools}
        tool_definitions = [t.tool_definition for t in tools]
        input = [{"role": "user", "content": user_prompt}]
        for _ in range(max_rounds):
            response = self.client.responses.create(
                model=use_model,
                input=input,
                tools=tool_definitions,
                **({"instructions": system_prompt} if system_prompt else {}),
                **kwargs,
            )
            function_calls = [item for item in response.output if item.type == "function_call"]
            if not function_calls:
                return response.output_text
            input.extend(response.output)
            for call in function_calls:
                result = tool_map[call.name](**json.loads(call.arguments))
                input.append({
                    "type": "function_call_output",
                    "call_id": call.call_id,
                    "output": result if isinstance(result, str) else json.dumps(result),
                })
        raise ValueError(f"Tool calling did not complete within {max_rounds} rounds")

    def embed(self, content, model=None, **kwargs):
        use_model = model or self.model
        if not use_model:
            raise ValueError("Model not provided in constructor nor at completion")
        embedding = self.client.embeddings.create(model=use_model, input=content, encoding_format="float", **kwargs)
        return embedding.data[0].embedding



"""
# tools for getting TLS certs

## usage:

```py
from a5.web import tls
bundle = tls.new_cert_bundle(domain, email, hosted_zone_id, staging=True)
with open("tls.json", "w") as f:
    f.write(tls.serialize_bundle(bundle))
```

```py
from a5.web import tls
with open("tls.json") as f:
    bundle = tls.deserialize_bundle(f.read())
nbundle = tls.renew_cert_bundle(bundle)
```
"""

import json
import logging
from pathlib import Path

import pico_acme
from pico_acme import route53

logger = logging.getLogger(__name__)


def new_cert_bundle(domain, account_email, route53_hosted_zone_id, staging=False):
    logger.info(f"Registering account")
    acme_client = pico_acme.register_account(account_email, agree_tos=True, staging=staging)
    logger.info(f"Creating key and CSR")
    key_pem_bin = pico_acme.make_key()
    csr_pem = pico_acme.make_csr(key_pem_bin, [domain])
    upsert, clean = route53.route53_upsert_cleanup(route53_hosted_zone_id)
    logger.info(f"Performing DNS01 challenge")
    fullchain_pem = pico_acme.perform_dns01(acme_client, domain, csr_pem, upsert, clean)
    logger.info(f"Success")
    return (
        domain,
        route53_hosted_zone_id,
        pico_acme.serialize_account(acme_client),
        key_pem_bin.decode("ascii"),
        fullchain_pem,
        staging,
    )


def renew_cert_bundle(bundle):
    domain, route53_hosted_zone_id, account_ser, key_pem, fullchain_pem, staging = bundle
    logger.info(f"Loading account")
    acme_client = pico_acme.deserialize_account(account_ser, staging=staging)
    logger.info(f"Loading key and creating CSR")
    csr_pem = pico_acme.make_csr(key_pem.encode("ascii"), [domain])
    upsert, clean = route53.route53_upsert_cleanup(route53_hosted_zone_id)
    logger.info(f"Performing DNS01 challenge for renewal")
    fullchain_pem = pico_acme.perform_dns01(acme_client, domain, csr_pem, upsert, clean)
    logger.info(f"Success")
    return (
        domain,
        route53_hosted_zone_id,
        pico_acme.serialize_account(acme_client),
        key_pem,
        fullchain_pem,
        staging,
    )


def maybe_renew_bundle(bundle):
    domain, route53_hosted_zone_id, account_ser, key_pem, fullchain_pem, staging = bundle
    if pico_acme.should_renew(fullchain_pem):
        bundle = renew_cert_bundle(bundle)
    return bundle


def serialize_bundle(bundle):
    return json.dumps(bundle)


def deserialize_bundle(data):
    return json.loads(data)


def get_key_and_cert(bundle):
    domain, route53_hosted_zone_id, account_ser, key_pem, fullchain_pem, staging = bundle
    return key_pem, fullchain_pem


def setup_tls_certs(storage_folder, domain, account_email, route53_hosted_zone_id, staging=False):
    """
    Sets up TLS. Gets new certs/renews as required. Returns paths to key and cert
    """
    storage = Path(storage_folder)
    storage.mkdir(parents=True, exist_ok=True)
    bundle_path = storage / "bundle.json"
    key_path = storage / "key.pem"
    fullchain_path = storage / "cert.pem"
    if bundle_path.exists() and key_path.exists() and fullchain_path.exists():
        # everything exists, just check if we need to renew
        bundle = deserialize_bundle(bundle_path.read_text())
        bundle = maybe_renew_bundle(bundle)
    else:
        bundle = new_cert_bundle(domain, account_email, route53_hosted_zone_id, staging=staging)
    bundle_path.write_text(serialize_bundle(bundle))
    key, chain = get_key_and_cert(bundle)
    key_path.write_text(key)
    fullchain_path.write_text(chain)
    return key_path, fullchain_path



from base64 import urlsafe_b64decode, urlsafe_b64encode
from typing import Optional

import nacl.pwhash
from nacl.bindings import crypto_aead
from nacl.bindings.crypto_generichash import generichash_blake2b_salt_personal
from nacl.bindings.utils import sodium_memcmp
from nacl.exceptions import InvalidkeyError
from nacl.hashlib import blake2b
from nacl.utils import random as random_bytes


def b64encode(data: bytes) -> str:
    return urlsafe_b64encode(data).decode("ascii")


def b64decode(data: str) -> bytes:
    return urlsafe_b64decode(data)


def urlsafe_random_bytes(length=32) -> str:
    return b64encode(random_bytes(length))


def urlsafe_secure_token():
    """
    A cryptographically secure random token that can be put in a URL
    """
    return urlsafe_random_bytes(32)


def cookiesafe_secure_token():
    return random_hex(32)


def hash_password(password: str):
    return nacl.pwhash.str(password.encode("utf-8"))


def verify_password(hashed: bytes, password: str):
    try:
        correct = nacl.pwhash.verify(hashed, password.encode("utf-8"))
        return correct
    except InvalidkeyError:
        return False


def random_hex(length=32):
    """
    Length in binary
    """
    return random_bytes(length).hex()


def compute_hash(data: bytes):
    return blake2b(data).hexdigest()


def secure_compare_bytes(val1: bytes, val2: bytes):
    return sodium_memcmp(val1, val2)


def secure_compare(val1: str, val2: str):
    return sodium_memcmp(val1.encode("utf8"), val2.encode("utf8"))


def generate_hash_signature(message: bytes, key: bytes) -> bytes:
    """
    Computes a blake2b keyed hash for the message.

    This can be used as a fast yet secure symmetric signature: by checking that
    the hashes agree, we can make sure the signature was generated by a party
    with knowledge of the key.
    """
    return generichash_blake2b_salt_personal(message, key=key, digest_size=32)


def verify_hash_signature(message: bytes, key: bytes, sig: bytes) -> bool:
    """
    Verifies a hash signature generated with generate_hash_signature.

    Returns true if the signature matches, otherwise false.
    """
    return secure_compare_bytes(sig, generate_hash_signature(message, key))


# AEAD: Authenticated Encryption with Associated Data

_aead_key_len = crypto_aead.crypto_aead_xchacha20poly1305_ietf_KEYBYTES
_aead_nonce_len = crypto_aead.crypto_aead_xchacha20poly1305_ietf_NPUBBYTES


def aead_generate_nonce():
    return random_bytes(_aead_nonce_len)


def aead_generate_key():
    return random_bytes(_aead_key_len)


def aead_encrypt(
    key: bytes,
    secret_data: bytes,
    plaintext_data: bytes = b"",
    nonce: Optional[bytes] = None,
) -> bytes:
    if not nonce:
        nonce = aead_generate_nonce()
    encrypted = crypto_aead.crypto_aead_xchacha20poly1305_ietf_encrypt(secret_data, plaintext_data, nonce, key)
    return nonce, encrypted


def aead_decrypt(key: bytes, nonce: bytes, encrypted_secret_data: bytes, plaintext_data: bytes = b"") -> bytes:
    return crypto_aead.crypto_aead_xchacha20poly1305_ietf_decrypt(encrypted_secret_data, plaintext_data, nonce, key)



import logging
from typing import Callable, Dict, Tuple, Union

from .http_status_codes import HTTP_STATUS_CODES
from .wsgi_server import run_wsgi_app

logger = logging.getLogger(__name__)


def basic_web_server(
    handler: Callable[[str, Dict], Tuple[Dict, Union[bytes, str]]],
    address,
    key_path=None,
    cert_path=None,
    workers=4,
):
    def app(environ, start_response):
        uri = environ["PATH_INFO"]
        logger.info(f"Got request for {uri}, calling handler")
        res = handler(uri, environ)
        if len(res) == 3:
            status, headers, response = res
        else:
            status = 200
            headers = {"Content-Type": "text/html"}
            response = res
        if isinstance(response, str):
            response = response.encode("utf8")
        if status is True:
            status = 200
        if isinstance(status, int):
            label = HTTP_STATUS_CODES.get(status, "unknown")
            status = f"{status} {label}"
        if isinstance(headers, dict):
            headers = list(headers.items())
        start_response(status, headers)
        logger.info(f"Responding to {uri}, with {status}, {len(response)} bytes, {len(headers)} headers")
        return [response]

    logger.info(f"Starting basic HTTP server on port {address}")
    run_wsgi_app(app, address, key_path=key_path, cert_path=cert_path, workers=workers)



import logging
from typing import Callable, Dict, Tuple, Union

from .wsgi_server import run_wsgi_app

logger = logging.getLogger(__name__)


def tiny_web_server(port, handler: Callable[[str], Tuple[Dict, Union[bytes, str]]]):
    def app(environ, start_response):
        if environ["REQUEST_METHOD"] != "GET":
            status = "400 Bad Request"
            headers = [("Content-type", "text/plain")]
            start_response(status, headers)
            return [b"Unsupported request."]
        uri = environ["PATH_INFO"]
        headers, response = handler(uri)
        if isinstance(response, str):
            response = response.encode("utf8")
        logger.info(f"Got request for {uri}, responding with {len(response)} bytes")
        start_response("200 OK", list(headers.items()))
        return [response]

    logger.info(f"Starting tiny HTTP server on port {port}")
    run_wsgi_app(app, f"0.0.0.0:{port}", workers=1)



import setuptools

with open("a5/version") as f:
    version = f.read().strip()

setuptools.setup(
    name="a5",
    version=version,
    author="Aapeli",
    description="a5",
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    package_data={"a5": ["version"]},
    python_requires=">=3.9",
    classifiers=[
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=["boto3", "gunicorn", "openai", "pico-acme", "pynacl", "tqdm"],
)



import logging

from .basic_web_server import basic_web_server

logger = logging.getLogger(__name__)


def make_static_handler(files):
    static = {url: (mime_type, data) for url, mime_type, data in files}
    logger.info(f"Generated {len(static)} static files")
    assert "error" in static, "Need a document called 'error'"

    def handler(uri, environ):
        code = 200
        if uri[0] != "/" or environ["REQUEST_METHOD"] != "GET":
            key = "error"
            code = 400
        else:
            key = uri[1:]
        if key not in static:
            key = "error"
            code = 404
        mime_type, data = static[key]
        return code, {"Content-Type": mime_type}, data

    return handler


def static_file_server(address_or_port, files, **kwargs):
    if isinstance(address_or_port, int):
        address_or_port = f"0.0.0.0:{address_or_port}"
    handler = make_static_handler(files)
    basic_web_server(handler, address_or_port, **kwargs)
